"""
This module provides event support for Osc things.
"""

import time
import socket
import traceback

from traceback import format_exc
from threading import Thread
from osc.oscAPI import *
from osc.OSC import *

class BaseEvent:
	def __init__(self):
		self.time = 0.0

class OscEvent(BaseEvent):

	def __init__(self,msg,tm=0.0):
		BaseEvent.__init__(self)
		self.oscmsg = msg

	def __str__(self):
		return self.to_xml()

	def to_xml(self):
		return '<event time=%.3f>%s</event>' % (self.time,self.oscmsg)

	def to_osc(self):
		return self.oscmsg

class OscRecipient:
	def __init__(self,addr,port):
		self.osc_addr = addr
		self.osc_port = int(port)
		self.osc_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

class OscMonitor:

	def __init__(self,addr,port):
		# if port is 0, one is chosen for us
		self.thread = OscThread(addr,port)
		self.thread.start()

	def getport(self):
		return self.thread.getport()

	def time_now(self):
		return time.time()

	def callback(self,f,data=None):
		return self.thread.callback(f,data)
		

class OscThread(Thread):

	def __init__(self,addr,port):
		Thread.__init__(self)
		self.addr = addr
		self.port = int(port)
		self.setDaemon(True)
		self.callback_func = None
		self.callback_data = None

		self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		self.sock.bind((self.addr,self.port))
		if self.port == 0:
			sockname = self.sock.getsockname()
			self.port = sockname[1]
		# addressManager = OSC.CallbackManager()
		# addressManager.add(self.handle_register,"/registerclient")
		# addressManager.add(self.handle_unregister,"/unregisterclient")

	def getport(self):
		return self.port

	def callback(self,f,data):
		self.callback_func = f
		self.callback_data = data

	def run(self):
		print "Listening for OSC on addr =",self.addr," port =",self.port
		while True:
			pair = self.sock.recvfrom(1024)
			try:
				t = decodeOSC(pair[0])
				# print "OSC received = ",t
				if self.callback_func != None:
					e = OscEvent(t)
					self.callback_func(e,self.callback_data)
				# addressManager.handle(pair[0])
			except:
				print "Unable to decode OSC message - %s"%(format_exc())

